import os
import cv2
import numpy as np
import gpio
import time
import config

def get_setting(key):
    return config.get_setting('imageProcessing', key).strip('""')

def get_setting_int(key):
    return int(get_setting(key))

def find_source_within_tmpl(blockNum, src_file, tmpl_file, tmpl_img_gray_resized, results_dir, scale, beeper_enabled):
    # Purpose:
    #   Find the source (individual numbered block) within the template (the mixed-up puzzle) and returns
    #   the results. Also creates a result image showing the match or matches. Lastly, optionally emits
    #   a short beep if at least one match is found.
    #   
    # Parameters:
    #   blockNum (integer): An integer between 1 and 15.
    #   src_file (string): The individual numbered block image. Relative to the current working directory.
    #   tmpl_file (string): Path/filename of template image (mixed-up puzzle). Relative to the current working directory.
    #   tmpl_img_gray_resized (bytes): The mixed-up puzzle image (usually from the camera), gray and resized.
    #   results_dir (string): Where to place the results. Relative to the current working directory.
    #   scale (float): How much to scale down the image to speed up the processing. Best when the image is high resolution. Values as low as 0.1 have been used successfully. 
    #   beeper_enabled (boolean): Should we beep.
    # Returns (list of tuples):
    #   Zero or more identified matches. Ideally these are closely positioned.
    #   Each tuple contains the top-left X-Y coordinates and the scale value at which it was found.
    
    # Load the source image (numbered block), convert to grayscale and get height/width info. Quick operations.
    source_img = cv2.imread(src_file)
    source_img_gray = cv2.cvtColor(source_img, cv2.COLOR_BGR2GRAY)
    source_img_gray_resized = cv2.resize(source_img_gray, None, fx=scale, fy=scale)
    source_height, source_width = source_img_gray_resized.shape[:2]
    
    thresholds = [0.95, 0.90, 0.85, 0.80, 0.775, 0.75, 0.725, 0.70]
    matches = []  # List to store the matched locations

    # Perform template matching. This line is where most of the time is spent.
    result = cv2.matchTemplate(source_img_gray_resized, tmpl_img_gray_resized, cv2.TM_CCOEFF_NORMED)

    # Select the results with the best threshold
    for threshold in thresholds:
        loc = np.where(result >= threshold)

        # Add the matched locations at this scale to the list
        for pt in zip(*loc[::-1]):
            matches.append((pt[0], pt[1]))

        if (len(matches) > 0):
            break

    if (len(matches) == 0):
        raise Exception(f"No results for block {blockNum}")

    # Draw rectangles around the matched regions on the source image
    matched_image = cv2.imread(tmpl_file)
    template_height, template_width = tmpl_img_gray_resized.shape[:2]
    for pt in matches:
        top = int(pt[0] / scale)
        left = int(pt[1] / scale)
        top_left = (top, left)
        bottom_right = (top + int(source_height / scale), left + int(source_width / scale))
        cv2.rectangle(matched_image, top_left, bottom_right, (0, 255, 0), 2)

    # Save the matched image with the drawn rectangles
    cv2.imwrite(os.path.join(results_dir, 'matched_' + str(blockNum) + '.jpg'), matched_image)
    
    beeper_enabled and gpio.beeper_short_beep()
        
    block_min_x = min(x for x, _ in matches)
    block_max_x = max(x for x, _ in matches)
    block_min_y = min(y for _, y in matches)
    block_max_y = max(y for _, y in matches)
    if (block_max_x - block_min_x > source_width or block_max_y - block_min_y > source_height): 
        print(f"It appears block {blockNum} can't be pinned to a general spot. Please review the images in folder {results_dir}.")
    
    #print('Block number ' + str(blockNum) + ' has ' + str(len(matches)) + ' matches, final threshold = ' + str(threshold))
    #for pt in matches:
    #    print(pt)
    
    return matches

def calc_block_x_y_avg_loc(allMatches):
    # Calculate average X and average Y values for each key using dictionary comprehension
    averages = {
        key: (sum(x for x, _ in matches) / len(matches), sum(y for _, y in matches) / len(matches))
        for key, matches in allMatches.items()
    }
    
    return averages

def calc_block_segments(sorted_blocks):
    # Determine which of the four rows the blocks belong to
    AVG_POS_DIFF_SCALE = 1.1
    returnValue = [[],[],[],[]]
    idx = 0
    previous_avg_pos = sorted_blocks[0][0]
    for current_avg_pos, key in sorted_blocks:
        if (AVG_POS_DIFF_SCALE * previous_avg_pos < current_avg_pos):
            idx += 1
        if (idx > 3):
            raise Exception(f"Unable to sort the blocks into just 4 rows. Possibly an image quality issue.")
        previous_avg_pos = current_avg_pos
        # print(f"Index {idx}, Key {key}")
        returnValue[idx].append(key)

    return returnValue

def process(tmpl_file, case_name, results_dir, results_file_tmpl, blocks_dir, scale, beeper_enabled, callback):
    # Purpose:
    #   Find the template image within the source image
    # Parameters:
    #   tmpl_file (string): Path/filename of template image (full puzzle). Absolute path. 
    #   case_name (string):
    #   results_dir (string):
    #   results_file_tmpl (string): 
    #   blocks_dir (string):
    #   scale (float): 
    #   template_img (string):
    #   beeper_enabled (boolean):
    #   callback (function):
    # Returns (list of integers): The numbered blocks in the order found in the template.
    
    if not os.path.exists(tmpl_file):
        raise Exception(f"Puzzle file not found: {tmpl_file}")
    
    # !!!! need to pass this   # Master07_01_23 = 1.0, Master07_02_23 = 1.0,  testpuzzle6_11_23 = 1.1
    results_dir = os.path.join(results_dir, case_name)
    tmpl_img = cv2.imread(tmpl_file)
    tmpl_img_gray = cv2.cvtColor(tmpl_img, cv2.COLOR_BGR2GRAY)
    tmpl_img_gray_resized = cv2.resize(tmpl_img_gray, None, fx=scale, fy=scale)
    allMatches = {}
    
    print(f"Placing results in {results_dir}")    
    print(f"Using numbered blocks at {os.path.join(blocks_dir, case_name, results_file_tmpl % 'X')}")
    
    for i in range(1, 16):
        if gpio.is_reset_pressed():
            raise Exception("user initiated abort")
        time.sleep(get_setting_int('BLOCK_PROCESSING_DELAY'))
        if gpio.is_reset_pressed():
            raise Exception("User initiated abort")
        
        block_num_file = os.path.join(blocks_dir, case_name, results_file_tmpl % str(i))
        if not os.path.exists(block_num_file):
            raise Exception(f"Block number file not found: {block_num_file}")
        
        matches = find_source_within_tmpl(i, block_num_file, tmpl_file, tmpl_img_gray_resized, results_dir, scale, beeper_enabled)
        allMatches[i] = matches
        
        four_by_four = convert_results_to_4x4(allMatches)
        callback(four_by_four)
        print(f"Interim answer {i}/15: {four_by_four}")

    return four_by_four
    
def convert_results_to_4x4(allMatches):
    
    blocks_x_y_avg_loc = calc_block_x_y_avg_loc(allMatches)

    # for key, values in blocks_x_y_avg_loc.items():
    #     average_x, average_y = values
    #     print(f"Key: {key}")
    #     print(f"Average X: {average_x}")
    #     print(f"Average Y: {average_y}")
    #     print()

    # Sort on block x and y values
    sorted_blocks_x = sorted([(avg_x, key) for key, (avg_x, _) in blocks_x_y_avg_loc.items()])
    sorted_blocks_y = sorted([(avg_y, key) for key, (_, avg_y) in blocks_x_y_avg_loc.items()])

    #for avg_x, key in sorted_blocks_x:
    #    print(f"Average X: {avg_x}, Key: {key}")

    #for avg_y, key in sorted_blocks_y:
    #    print(f"Average Y: {avg_y}, Key: {key}")

    x_segments = calc_block_segments(sorted_blocks_x)
    #print(x_segments)

    y_segments = calc_block_segments(sorted_blocks_y)
    #print(y_segments)

    x_map_block_no_to_grp = {num: idx for idx, inner_list in enumerate(x_segments) for num in inner_list}
    #print(x_map_block_no_to_grp)

    y_map_block_no_to_grp = {num: idx for idx, inner_list in enumerate(y_segments) for num in inner_list}
    #print(y_map_block_no_to_grp)

    result = [16 for _ in range(16)]

    for block_no, col in x_map_block_no_to_grp.items():
        idx = y_map_block_no_to_grp[block_no] * 4 + col
        result[idx] = block_no

    #print(result)
    return result
