#Python Program for PuzzleBot
#Date: 8/17/23
#Programmer: Matt Borack
#PuzzleBot: Dave Goodsell
#Processor: Raspberry Pi 3B+
#Modules:
#main.py
#gpio.py
#camera.py
#imageProcessing.py
#lcd.py
#serialPort.py
#config.py

import os
import time
import datetime
import logging
import signal
import sys
import enum

import camera
import lcd
import gpio
import serialPort
import imageProcessing

def handle_hup_signal(signum, frame):
    print("Received HUP")

signal.signal(signal.SIGHUP, handle_hup_signal)

import config

def get_setting(key):
    return config.get_setting('main', key).strip('"')

def get_setting_int(key):
    return int(get_setting(key))

def get_setting_bool(key):
    return get_setting(key) == 'True'

def send_update(msg):
    print(f"{datetime.datetime.now()}: {msg}")

def adj_start():
    
    send_update("Starting adjustment mode. Flip the switch to Run mode when done.")
    change_state_to_adjusting()
    
    try:
        gpio.photo_light_on()
        camera.start_camera_preview()        
    except BaseException as e:
        camera.end_camera_preview()
        gpio.photo_light_off()
        
        gpio.reset_all()
        logger.exception(e)
        change_state_to_error()
    finally:
        last_on_go = time.time()
    
def adj_end():
    
    try:
        camera.end_camera_preview()
        gpio.photo_light_off()
    except BaseException as e:
        gpio.reset_all()
        logger.exception(e)
        change_state_to_error()
    
def on_go(channel):
    
    # Deal with bounce
    global last_on_go
    if (time.time() - last_on_go < 1):
        return
    
    send_update("Go signal detected")
    gpio.led_ready_off()
    change_state_to_processing()
    
    try:
        LCD_ENABLED and lcd.clear(ser)
        image_file = os.path.join(DIR_PUZZLE_IMAGES, DIR_CASE_NAME, 'camera.jpg') 
                
        gpio.photo_light_on()        
        camera.take_pic(image_file, True, CAMERA_DELAY)
        gpio.photo_light_off()
                
        override_camera_with = get_setting('OVERRIDE_CAMERA_WITH')
        if (override_camera_with != ''):
            image_file = os.path.join(DIR_PUZZLE_IMAGES, DIR_CASE_NAME, override_camera_with)
            print("Overriding camera due to config key 'OVERRIDE_CAMERA_WITH'")
        print(f"Using puzzle image {image_file}")
        print(f"To reset, press and hold the Reset button until the code responds.")
        result = imageProcessing.process(image_file, DIR_CASE_NAME, DIR_RESULTS, RESULTS_FILE_TMPL, DIR_PUZZLE_BLOCK_IMAGES, SCALE, BEEPER_ENABLED, on_go_cb)
        #result = [16, 15, 14, 12, 13, 4, 7, 9, 8, 3, 10, 5, 6, 1, 11, 2] # !!!
        
        serial_result = serialPort.convert_result(result, True)
        LCD_ENABLED and lcd.display_result(ser, serial_result)
        
        if BEEPER_ENABLED:
            for i in range(1, 4):
                gpio.beeper_short_beep()
                time.sleep(1/10)
                        
        send_update("Finished processing")
        change_state_to_success()
    except BaseException as e:
        BEEPER_ENABLED and gpio.beeper_on(2)
        gpio.reset_all()
        logger.exception(e)
        change_state_to_error()
    finally:
        print("------------------------------------------------------")    
        last_on_go = time.time()
    
def on_go_cb(result):
    serial_result = serialPort.convert_result(result, False)        
    LCD_ENABLED and lcd.display_result(ser, serial_result)    

def on_reset(channel):    
    send_update("Reset signal detected")
    change_state_to_ready()
        
def on_mode(channel):
    send_update("Mode signal detected.")
    adj_end()
    
def change_state_to_ready():
    send_update("State = Ready")
    global state
    state = 'READY'
    gpio.led_ready_on()

def change_state_to_processing():
    send_update("State = Processing")
    global state
    state = 'PROCESSING'
    gpio.led_processing_on()

def change_state_to_adjusting():
    send_update("State = Adjusting")
    global state
    state = 'ADJUSTING'
    gpio.led_processing_off()
    # Note: No LED light at this time to indicate when adjusting

def change_state_to_success():
    send_update("State = Success")
    global state
    state = 'SUCCESS'
    gpio.led_processing_on()

def change_state_to_error():
    send_update("State = Error. Press the Reset button to start over.")
    global state
    state = 'ERROR'
    gpio.led_error_on()

def create_dir_if_needed(dir_name, sub_dir_name):
    if not os.path.exists(dir_name):
        os.mkdir(dir_name)
    if not os.path.exists(os.path.join(dir_name, sub_dir_name)):
        os.mkdir(os.path.join(dir_name, sub_dir_name))


#STATES = enum('States', ['READY','PROCESSING','ADJUSTING','SUCCESS','ERROR'])

BEEPER_ENABLED = get_setting_bool('BEEPER_ENABLED') # Set to False when your family is sleeping :)
CAMERA_DELAY = get_setting_int('CAMERA_DELAY') # How long the light should be on before taking the picture
# Camera rotation - see the camera.py module
LCD_ENABLED = get_setting_bool('LCD_ENABLED') # Set to True when the LCD is connected to the serial port
DIR_RESULTS = get_setting('DIR_RESULTS') # Name of directory to place the result images. Relative to current directory.
DIR_PUZZLE_BLOCK_IMAGES = get_setting('DIR_PUZZLE_BLOCK_IMAGES') # Name of directory to find puzzle block images. Relative to current directory.
DIR_PUZZLE_IMAGES = get_setting('DIR_PUZZLE_IMAGES') # Name of directory to find (full) puzzle images. Relative to current directory.
DIR_CASE_NAME = get_setting('DIR_CASE_NAME') # Name of the sub directory to place/find the above images. Useful when testing.
SCALE = float(get_setting('SCALE_IMAGE_DOWN_TO'))
    # How much to scale down the image to speed up the processing. Best when the image is high resolution. Values as low as 0.1 have been used successfully. 
    # Assumption: The puzzle image and the numbered blocks are at the same scale.
RESULTS_FILE_TMPL = 'Block-%s.jpg'


global state
state = ''
logger = logging.Logger('')

create_dir_if_needed(DIR_RESULTS, DIR_CASE_NAME)
create_dir_if_needed(DIR_PUZZLE_BLOCK_IMAGES, DIR_CASE_NAME)
create_dir_if_needed(DIR_PUZZLE_IMAGES, DIR_CASE_NAME)

ser = serialPort.init()
gpio.init(on_go, on_reset, on_mode)
gpio.reset_all()

# Add a slight delay to cause a flickering effect. Useful for those cases where the Ready light
# is already on.
time.sleep(0.25)
change_state_to_ready()

# Helps with debouncing
global last_on_go
last_on_go = 0

if gpio.is_mode_adj():
    adj_start()

try:
    input("Press the ENTER key when not processing to break\n") # prevent program from exiting
finally:
    # Clean up
    gpio.reset_all()
    LCD_ENABLED and lcd.clear(ser)


